/*
  This file is part of CDO. CDO is a collection of Operators to manipulate and analyse Climate model Data.

  Author: Uwe Schulzweida

*/
#ifndef FIELD_H
#define FIELD_H

#include <variant>
#include <stdexcept>

#include <cstdio>

#include "varray.h"
#include "cdo_options.h"
#include "cdo_vlist.h"
#include "compare.h"


template<typename T>
class DataVec
{
public:
  DataVec() = default;
  DataVec(size_t size);

  Varray<T> vec;

  void resize(size_t count);
  void resize(size_t count, const T value);
  template<typename U>
  void resize(size_t count, const U value);
};

template<typename T>
DataVec<T>::DataVec(size_t size)
{
  varrayResize(vec, size);
}

// In the original version was a resize for double and a resizef for  float
template<typename T>
inline void DataVec<T>::resize(const size_t count)
{
  //m_count = count;
  //varrayResize(vec, m_count);
  varrayResize(vec, count);
  // if (!size)  size = m_count;
}

template<typename T>
inline void DataVec<T>::resize(const size_t count, const T value)
{
  //m_count = count;
  //varrayResizeInit(vec, m_count, value);
  varrayResizeInit(vec, count, value);
  //if (!size)  size = m_count;
}

template<typename T>
template<typename U>
inline void DataVec<T>::resize(const size_t count, const U value)
{
  //m_count = count;
  //varrayResizeInit(vec, m_count, static_cast<T>(value));
  varrayResizeInit(vec, count, static_cast<T>(value));
  // if (!size)  size = m_count;
}

using DataFloat = DataVec<float>;
using DataDouble = DataVec<double>;
class VarData;

#define LambdaVarData_ auto&

class VarData : public std::variant<DataDouble, DataFloat>
{
public:
  using _var_ = std::variant<DataDouble, DataFloat>;

public:
  // Overwrites of the std::variant<> constructors
  constexpr VarData() noexcept;
  VarData(VarData& other);

  template<class T>
  constexpr VarData(T&& t) noexcept;

  template<class T, class... Args>
  constexpr explicit VarData(std::in_place_type_t<T>, Args&&... args);

  template<class T, class U, class... Args>
  constexpr explicit VarData(std::in_place_type_t<T>, std::initializer_list<U> il, Args&&... args);

  template<std::size_t I, class... Args>
  constexpr explicit VarData(std::in_place_index_t<I>, Args&&... args);

  template< std::size_t I, class U, class... Args>
  constexpr explicit VarData(std::in_place_index_t<I>, std::initializer_list<U> il, Args&&... args);

  // Overwrites of the std::variant<> operator=
  VarData& operator=(const VarData& rhs);
  VarData& operator=(VarData&& rhs) noexcept;

  template<class T>
  VarData& operator=(T&& t) noexcept;

  // Self defined methods
  void resize(size_t count);

  template<typename U>
  void resize(size_t count, const U value);

  void setMissValue(double value);
  double getMissValue() const;

  bool empty() const;

  size_t getNumMiss() const;
  void setNumMiss(const size_t value);

  size_t getSize() const;
};

inline constexpr VarData::VarData() noexcept : _var_()
{
}

inline VarData::VarData(VarData& other) : _var_(other)
{
}

template< class T >
inline constexpr VarData::VarData( T&& t ) noexcept : _var_(std::move(t))
{
}

template<class T, class... Args>
inline constexpr VarData::VarData(std::in_place_type_t<T>, Args&&... args)
    : _var_(std::in_place_type_t<T>{}, args...)
{
}

template<class T, class U, class... Args>
inline constexpr VarData::VarData(std::in_place_type_t<T>, std::initializer_list<U> il, Args&&... args )
    : _var_(std::in_place_type_t<T>{}, il, args...)
{
}

template<std::size_t I, class... Args>
inline constexpr VarData::VarData(std::in_place_index_t<I>, Args&&... args)
    : _var_(std::in_place_index_t<I>{}, args...)
{
}

template<std::size_t I, class U, class... Args>
inline constexpr VarData::VarData(std::in_place_index_t<I>, std::initializer_list<U> il, Args&&... args)
    : _var_(std::in_place_index_t<I>{}, il, args...)
{
}

// Overwrites of the std::variant<> operator=
inline VarData& VarData::operator=(const VarData& rhs)
{
  _var_::operator=(rhs);
  return *this;
}

inline VarData& VarData::operator=(VarData&& rhs) noexcept
{
  _var_::operator=(std::move(rhs));
  return *this;
}

template< class T >
inline VarData& VarData::operator=(T&& t) noexcept
{
  _var_::operator=(std::move(t));
  return *this;
}

class // VField
#ifdef WARN_UNUSED
[[gnu::warn_unused]]
#endif
VField
{
public:
  MemType memType = MemType::Native;  // MemType::Float or MemType::Double

  int nwpv = 1;         // number of words per value; real:1  complex:2
  int grid = -1;

  size_t gridsize = 0;
  size_t size = 0;

  size_t nmiss = 0;
  double missval = 0;

  VarData vdata;
  Varray<double> weightv;

  VField() {}
  void init(const CdoVar &var);

private:
  size_t m_count = 0;
};


inline VarData init(const CdoVar& var)
{
  if (var.memType == MemType::Float)
    {
      return DataVec<float>{var.gridsize};
    }
  else if (var.memType == MemType::Double)
    {
      return DataVec<double>{var.gridsize};
    }
  else
    {
      throw std::runtime_error("Not supported type");
    }
}

inline void reInit(VarData &data, const CdoVar& var)
{
  if ((var.memType == MemType::Float && std::holds_alternative<DataDouble>(data)) ||
      (var.memType == MemType::Double && std::holds_alternative<DataFloat>(data)))
    {
      data = init(var); // The type has changed
    }
  else if (var.memType == MemType::Double && std::holds_alternative<DataDouble>(data))
    {
      std::get<DataDouble>(data).resize(var.gridsize * var.nwpv);
    }
  else if (var.memType == MemType::Float && std::holds_alternative<DataFloat>(data))
    {
      std::get<DataFloat>(data).resize(var.gridsize * var.nwpv);
    }
  else
    {
      throw std::runtime_error("Not supported type");
    }
}

inline void
VField::init(const CdoVar &var)
{
  nwpv = 1;
  grid = var.gridID;
  gridsize = var.gridsize;
  nmiss = 0;
  missval = var.missval;
  memType = var.memType;
  size = var.gridsize * var.nwpv;
  m_count = size;
  /*
  if (memType == MemType::Float)
    vec_f.resize(size);
  else
    vec_d.resize(size);
  */

  reInit(vdata, var);
}

template<typename T>
struct FieldX_t {
  using typeFloat = void;
  using typeDouble = void;
};

template<>
struct FieldX_t<VarData>
{
  using typeFloat = DataFloat;
  using typeDouble = DataDouble;
};

template<>
struct FieldX_t<const VarData>
{
  using typeFloat = DataFloat;
  using typeDouble = DataDouble;
};
/*
inline bool vdata_have_same_type(const VarData& vdata1, const VarData& vdata2)
{
  if (std::holds_alternative<DataFloat>(vdata1) && std::holds_alternative<DataFloat>(vdata2))
    {
      return true;
    }
  else if(std::holds_alternative<DataDouble>(vdata1) && std::holds_alternative<DataDouble>(vdata2))
    {
      return true;
    }
  else
    {
      return false;
    }
}
*/
template<typename T>
inline constexpr bool check_T_is_vdata()
{
  constexpr bool is_vdata = std::is_same<T, VarData>::value;
  constexpr bool is_const_vdata = std::is_same<T, const VarData>::value;

  return is_vdata || is_const_vdata;
}

template<typename T>
inline constexpr bool check_T_is_vdata_t()
{
  constexpr bool is_vdata = check_T_is_vdata<T>();

  return is_vdata;
}


template<typename Function, typename VarData_T_, typename... ARGS>
inline auto
vdata_operation(Function f, VarData_T_& vdata, ARGS &... args) 
-> decltype(f(std::declval<typename FieldX_t<VarData_T_>::typeFloat&>(), args...));

// clang-format off
template<typename FunctionFloat, typename FunctionDouble, typename VarData_T_, typename... ARGS>
inline auto
vdata_operation(FunctionFloat functionFloat, FunctionDouble functionDouble, VarData_T_& vdata, ARGS &... args) 
-> decltype(functionFloat(std::declval<typename FieldX_t<VarData_T_>::typeFloat&>(), args...)) {

  using typeFloat  = typename FieldX_t<VarData_T_>::typeFloat;
  using typeDouble = typename FieldX_t<VarData_T_>::typeDouble;

  // check that functionFloat and functionDouble has the same signatur and specialy the same return type
  using returnTypeFunctionFloat = decltype(functionFloat(std::declval<typeFloat&>(), args...));
  using returnTypeFunctionDouble = decltype(functionDouble(std::declval<typeDouble&>(), args...));
  static_assert(std::is_same<returnTypeFunctionFloat, returnTypeFunctionDouble>::value,
                "functionFloat and functionDouble has not the same return types");

  // Fail fast if vdata is not from type `VarData` or `const VarData`
  static_assert(check_T_is_vdata_t<VarData_T_>(), "data must be from type 'VarData' or 'const VarData'.");

  if (std::holds_alternative<typeFloat>(vdata))
    {
      return functionFloat(std::get<typeFloat>(vdata), args...);
    }
  else if(std::holds_alternative<typeDouble>(vdata))
    {
      return functionDouble(std::get<typeDouble>(vdata), args...);
    }
  else
    {
      throw std::runtime_error("Not supported Type");
    }
}
//clang-format on

// clang-format off
template<typename Function, typename VarData_T_, typename... ARGS>
inline auto
vdata_operation(Function f, VarData_T_& vdata, ARGS &... args) 
-> decltype(f(std::declval<typename FieldX_t<VarData_T_>::typeFloat&>(), args...))
{
  // Fail fast if vdata is not from type `VarData` or `const VarData`
  static_assert(check_T_is_vdata_t<VarData_T_>(), "field must be from type 'VarData' or 'const VarData'");

  return vdata_operation(f, f, vdata, args...);
}
// clang-format on

// clang-format off
template<typename Function, typename VarData_T1_, typename VarData_T2_, typename... ARGS>
inline auto
vdata_operation2(Function f, VarData_T1_& vdata1, VarData_T2_& vdata2, ARGS... args) 
-> decltype(f(
              std::declval<typename FieldX_t<VarData_T1_>::typeFloat&>(), 
              std::declval<typename FieldX_t<VarData_T2_>::typeFloat&>(),
              args...))
{
  using typeFloat  = typename FieldX_t<VarData_T1_>::typeFloat;
  using typeDouble = typename FieldX_t<VarData_T1_>::typeDouble;

  // check that functionFloat and functionDouble has the same signatur and specialy the same return type
  using returnTypeFunctionFloatFloat = decltype(f(std::declval<typeFloat&>(), std::declval<typeFloat&>(), args...));
  using returnTypeFunctionFloatDouble = decltype(f(std::declval<typeFloat&>(), std::declval<typeDouble&>(), args...));
  using returnTypeFunctionDoubleDouble = decltype(f(std::declval<typeDouble&>(), std::declval<typeDouble&>(), args...));
  using returnTypeFunctionDoubleFloat = decltype(f(std::declval<typeDouble&>(), std::declval<typeFloat&>(), args...));
  static_assert(
    std::is_same<returnTypeFunctionFloatFloat, returnTypeFunctionFloatDouble>::value &&
    std::is_same<returnTypeFunctionFloatFloat, returnTypeFunctionDoubleDouble>::value &&
    std::is_same<returnTypeFunctionFloatFloat, returnTypeFunctionDoubleFloat>::value,
  "f needs the same return types for all input types");

  // Fail fast if vdata is not from type `VarData` or `const VarData`
  static_assert(check_T_is_vdata_t<VarData_T1_>(), "vdata1 must be from type 'VarData' or 'const VarData' or 'Field3D_t' or 'const Field3D_t'");
  static_assert(check_T_is_vdata_t<VarData_T2_>(), "vdata2 must be from type 'VarData' or 'const VarData' or 'Field3D_t' or 'const Field3D_t'");

  if (std::holds_alternative<typeFloat>(vdata1))
    {
      if      (std::holds_alternative<typeFloat>(vdata2))  return f(std::get<typeFloat>(vdata1), std::get<typeFloat>(vdata2),  args...);
      else if (std::holds_alternative<typeDouble>(vdata2)) return f(std::get<typeFloat>(vdata1), std::get<typeDouble>(vdata2),  args...);
      else throw std::runtime_error("Not supported Type of vdata2");
    }
  else if (std::holds_alternative<typeDouble>(vdata1))
    {
      if      (std::holds_alternative<typeFloat>(vdata2))  return f(std::get<typeDouble>(vdata1), std::get<typeFloat>(vdata2),  args...);
      else if (std::holds_alternative<typeDouble>(vdata2)) return f(std::get<typeDouble>(vdata1), std::get<typeDouble>(vdata2),  args...);
      else throw std::runtime_error("Not supported Type of vdata2");
    }
  else
    {
      throw std::runtime_error("Not supported Type of vdata1");
    }
}
//clang-format on

inline void VarData::resize(size_t count)
{
  vdata_operation([=](LambdaVarData_ vdata){vdata.resize(count);}, *this);
}

template<typename U>
inline void VarData::resize(size_t count, const U value)
{
  vdata_operation([=](LambdaVarData_ vdata){vdata.resize(count, value);}, *this);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// clang-format off
const auto memtype_is_float_float   = [](auto a, auto b) noexcept { return (a == MemType::Float  && b == MemType::Float); };
const auto memtype_is_double_float  = [](auto a, auto b) noexcept { return (a == MemType::Double && b == MemType::Float); };
const auto memtype_is_float_double  = [](auto a, auto b) noexcept { return (a == MemType::Float  && b == MemType::Double); };
const auto memtype_is_double_double = [](auto a, auto b) noexcept { return (a == MemType::Double && b == MemType::Double); };
// clang-format on

enum field_flag
{
  FIELD_VEC = 2,  // allocated memory
  FIELD_FLT = 4,  // 32-bit float
  FIELD_DBL = 8,  // 64-bit float
  FIELD_NAT = 16, // native: 32-bit float for 32-bit float data, otherweise 64-bit float
};

class // Field
#ifdef WARN_UNUSED
[[gnu::warn_unused]]
#endif
Field
{
public:
  int fpeRaised = 0;
  MemType memType = MemType::Native;  // MemType::Float or MemType::Double

  int nwpv = 1;         // number of words per value; real:1  complex:2
  int grid = -1;

  size_t gridsize = 0;
  size_t size = 0;
  size_t nsamp = 0;

  size_t nmiss = 0;
  double missval = 0;

  Varray<float> vec_f;
  Varray<double> vec_d;
  Varray<double> weightv;

  Field() {}
  void init(const CdoVar &var);
  void resize(size_t count);
  void resize(size_t count, double value);
  void resizef(size_t count);
  void resizef(size_t count, float value);
  bool empty() const;
  void check_gridsize() const;

  bool hasData() const
  {
    return (memType == MemType::Float) ? !vec_f.empty() : !vec_d.empty();
  }

private:
  size_t m_count = 0;
};

class // Field3D
#ifdef WARN_UNUSED
[[gnu::warn_unused]]
#endif
Field3D : public Field
{
public:
  size_t nlevels = 0;

  Field3D() {}
  void init(const CdoVar &var);
};

struct RecordInfo
{
  int varID = 0;
  int levelID = 0;
  bool lconst = false;
};

using FieldVector = std::vector<Field>;
using FieldVector2D = std::vector<FieldVector>;
using FieldVector3D = std::vector<FieldVector2D>;

using Field3DVector = std::vector<Field3D>;

void field_fill(Field &field, double value);
void field_ncopy(size_t n, const Field &fieldIn, Field &fieldOut);
void field_copy(const Field &fieldIn, Field &fieldOut);
void field_copy(const Field3D &fieldIn, Field3D &fieldOut);
void field_copy(const Field3D &fieldIn, int levelID, Field &fieldOut);
void field_add(Field &field1, const Field3D &field2, int levelID);
size_t field_num_miss(const Field &field);
size_t field_num_mv(Field &field);
MinMax field_min_max(const Field &field);

template <class UnaryOperation>
void field_transform(const Field &fieldIn, Field &fieldOut, UnaryOperation unary_op)
{
  if (fieldIn.memType == MemType::Float && fieldOut.memType == MemType::Float)
    varray_transform(fieldIn.vec_f, fieldOut.vec_f, unary_op);
  else
    varray_transform(fieldIn.vec_d, fieldOut.vec_d, unary_op);
}

#endif /* FIELD_H */
